{-# LANGUAGE LambdaCase #-}

module PUBMED.Parser where

import Control.Lens ((^?), (^.), (^..), only, to, ix, prism', Prism', Traversal')
import Data.Maybe (catMaybes, fromMaybe)
import Data.Time (UTCTime(..))
import Data.Time.Segment (jour)
import Prelude hiding (head)
import Text.Taggy (Node(..))
import qualified Data.ByteString.Lazy as DBL
import qualified Data.Text as T
import qualified Data.Text.Lazy as TL
import qualified Data.Text.Lazy.IO as TLIO
import Panic (panic)
import qualified Text.Read as TR
import qualified Text.Taggy.Lens as TTL

namedEl name = TTL.elements . TTL.named (only name)

contentWithChildren :: Prism' Node T.Text
contentWithChildren = prism' NodeContent $ \case { NodeContent c -> Just c
                                                 ; NodeElement e -> Just $ e ^. TTL.children . traverse . contentWithChildren }

deepContent = TTL.children . traverse . contentWithChildren

parseDocIds :: TL.Text -> [Integer]
parseDocIds txt = map parseId parsed
  where
    parsed = txt ^.. TTL.html . TTL.allNamed (only "eSearchResult") . namedEl "IdList" . namedEl "Id" . TTL.contents
    parseId s = case (TR.readMaybe (T.unpack s) :: Maybe Integer) of
      Nothing -> panic $ "Can't read doc id from: " <> (T.unpack s)
      Just cnt -> cnt

parseDocCount :: TL.Text -> Maybe Integer
parseDocCount txt = TR.readMaybe $ T.unpack $ txt ^. TTL.html . TTL.allNamed (only "eSearchResult") . namedEl "Count" . TTL.contents

data PubMed =
  PubMed { pubmed_id      :: Int
         , pubmed_article :: PubMedArticle
         , pubmed_date    :: PubMedDate
         } deriving Show

data PubMedArticle =
     PubMedArticle { pubmed_title   :: Maybe T.Text
                   , pubmed_journal :: Maybe T.Text
                   , pubmed_abstract :: [T.Text]
                   , pubmed_authors :: [Author]
                   }
                   deriving (Show)

data Author =
  Author { lastName :: Maybe T.Text
         , foreName :: Maybe T.Text
         , affiliation :: Maybe T.Text
         }
  deriving (Show)

data PubMedDate =
     PubMedDate { pubmedDate_date     :: UTCTime
                , pubmedDate_year     :: Integer
                , pubmedDate_month    :: Int
                , pubmedDate_day      :: Int
                } deriving (Show)


parsePubMedFile fp = do
  input <- TLIO.readFile fp
  pure $ parsePubMed input

parsePubMed :: TL.Text -> [PubMed]
parsePubMed txt = catMaybes $ txt ^.. pubmedArticle . to pubMed
  where
    pubmedArticle = TTL.html . TTL.allNamed (only "PubmedArticleSet") . TTL.allNamed (only "PubmedArticle")
    pubMed el = PubMed <$> el ^? articleId . to pubMedId
                       <*> el ^? article . to pubMedArticle
                       <*> el ^? pubDate . to pubMedDate
      where
        medline :: Traversal' TTL.Element TTL.Element
        medline = namedEl "MedlineCitation"
        article = medline . namedEl "Article"
        articleId = medline . namedEl "PMID"
        pubDate = namedEl "PubmedData" . namedEl "History" . namedEl "PubMedPubDate" . TTL.attributed (ix "PubStatus" . only "pubmed")
    pubMedId el = case (TR.readMaybe $ T.unpack $ el ^. TTL.contents) of
      Nothing -> panic $ "Cannot parse id: " <> (T.unpack $ el ^. TTL.contents)
      Just id -> id
    pubMedDate el = PubMedDate { pubmedDate_date = jour y m d
                               , pubmedDate_year = y
                               , pubmedDate_month = m
                               , pubmedDate_day = d }
      where
        y = fromMaybe 1 $ TR.readMaybe $ T.unpack $ el ^. namedEl "Year" . TTL.contents
        m = fromMaybe 1 $ TR.readMaybe $ T.unpack $ el ^. namedEl "Month" . TTL.contents
        d = fromMaybe 1 $ TR.readMaybe $ T.unpack $ el ^. namedEl "Day" . TTL.contents
    pubMedArticle el = PubMedArticle { pubmed_title = Just $ el ^. title
                                     , pubmed_journal = el ^? journalTitle
                                     , pubmed_abstract = el ^.. abstract
                                     , pubmed_authors = el ^.. authors . to author }
      where
        journalTitle = namedEl "Journal" . namedEl "Title" . TTL.contents
        title = namedEl "ArticleTitle" . deepContent
        abstract = namedEl "Abstract" . TTL.allNamed (only "AbstractText") . to abstractContent
          where
            abstractContent el = el ^. deepContent
        authors = namedEl "AuthorList" . TTL.allNamed (only "Author")
    author el = Author { lastName = el ^? lastName
                       , foreName = el ^? firstName
                       , affiliation = el ^? affiliation }
      where
        firstName = namedEl "ForeName" . TTL.contents
        lastName = namedEl "LastName" . TTL.contents
        affiliation = namedEl "AffiliationInfo" . namedEl "Affiliation" . TTL.contents


pubMedData :: DBL.ByteString
pubMedData = mconcat
  [ "<?xml version=\"1.0\"?>\n"
  , "<!DOCTYPE PubmedArticleSet PUBLIC \"-//NLM//DTD PubMedArticle, 1st June 2018//EN\" \"https://dtd.nlm.nih.gov/ncbi/pubmed/out/pubmed_180601.dtd\">\n"
  , "<PubmedArticleSet>\n"
  , "<PubmedArticle>\n"
  , " <MedlineCitation Status=\"Publisher\" Owner=\"NLM\">\n"
  , "        <PMID Version=\"1\">30357468</PMID>\n"
  , "        <DateRevised>\n"
  , "           <Year>2018</Year>\n"
  , "        </DateRevised>\n"
  , "        <Article PubModel=\"Print-Electronic\">\n"
  , "          <Journal>\n"
  , "            <ISSN IssnType=\"Electronic\">1432-1076</ISSN>\n"
  , "            <Title>European journal of pediatrics</Title>\n"
  , "          </Journal>\n"
  , "          <ArticleTitle>Title of the Article</ArticleTitle>\n"
  , "          <ELocationID EIdType=\"doi\" ValidYN=\"Y\">10.1007/s00431-018-3270-3</ELocationID>\n"
  , "          <Abstract>\n"
  , "              <AbstractText>Abstract Text.</AbstractText>\n"
  , "          </Abstract>\n"
  , "          <AuthorList>\n"
  , "          </AuthorList>\n"
  , "        </Article>\n"
  , " </MedlineCitation>\n"
  , " <PubmedData>\n"
  , "   <History>\n"
  , "   </History>\n"
  , " </PubmedData>\n"
  , "</PubmedArticle>\n"
  , "</PubmedArticleSet>\n"
  ]
